package engine

import (
	"fmt"
	"math/big"
	"strings"

	"github.com/exchain/process/orderbook"
	"github.com/exchain/process/types"
	"github.com/exchain/process/utils"

	"github.com/ethereum/go-ethereum/common"
	"github.com/exchain/go-exchain/exchain"
	"github.com/exchain/go-exchain/exchain/chaindb"
	nebulav1 "github.com/exchain/go-exchain/exchain/protocol/gen/go/nebula/v1"
	"github.com/exchain/go-exchain/exchain/wrapper"
	"github.com/holiman/uint256"
)

func (e *Engine) ProcessPayload(block *nebulav1.Block) (exchain.ExecutionResult, error) {
	genesis := block.Header.Height == 0
	if !genesis {
		parent, err := e.chainDB.GetBlockByLabel(chaindb.ExChainBlockLatest)
		if err != nil {
			return exchain.ExecutionResult{}, err
		}
		if parent.Header.Height+1 != block.Header.Height {
			return exchain.ExecutionResult{}, fmt.Errorf("invalid block height")
		}
	}
	receipts, err := e.ProcessTx(block.Header, block.Transactions)
	if err != nil {
		return exchain.ExecutionResult{}, err
	}

	return exchain.ExecutionResult{
		Payload:  block,
		Receipts: receipts,
	}, nil
}

func (e *Engine) ProcessOrders(header *nebulav1.BlockHeader) (txs *nebulav1.TransactionList, receipts *nebulav1.TransactionReceiptList, err error) {
	txs = &nebulav1.TransactionList{}
	txs.Txs = make([]*nebulav1.Transaction, 0)
	receipts = &nebulav1.TransactionReceiptList{}
	receipts.Receipts = make([]*nebulav1.TransactionReceipt, 0)

outer:
	for i := 0; i < 100 && len(txs.Txs) < 100; i++ {
		select {
		case tx := <-e.txQueue:
			if tx == nil {
				break
			}
			txs.Txs = append(txs.Txs, tx)
		default:
			break outer
		}
	}

	for _, tx := range txs.Txs {
		wtx := wrapper.NewTxWrapper(tx)
		switch tx.TxType {
		case nebulav1.TxType_LimitTx:
			receipts.Receipts = append(receipts.Receipts, &nebulav1.TransactionReceipt{
				Hash:        wtx.Hash().Bytes(),
				TxType:      tx.TxType,
				Success:     true,
				BlockHeight: header.Height,
				Timestamp:   header.Timestamp,
				Content: &nebulav1.TransactionReceipt_LimitR{
					LimitR: &nebulav1.LimitOrderReceipt{},
				},
			})
		case nebulav1.TxType_MarketTx:
			receipts.Receipts = append(receipts.Receipts, &nebulav1.TransactionReceipt{
				Hash:        wtx.Hash().Bytes(),
				TxType:      tx.TxType,
				Success:     true,
				BlockHeight: header.Height,
				Timestamp:   header.Timestamp,
				Content: &nebulav1.TransactionReceipt_MarketR{
					MarketR: &nebulav1.MarketOrderReceipt{},
				},
			})
		case nebulav1.TxType_CancelTx:
			receipts.Receipts = append(receipts.Receipts, &nebulav1.TransactionReceipt{
				Hash:        wtx.Hash().Bytes(),
				TxType:      tx.TxType,
				Success:     true,
				BlockHeight: header.Height,
				Timestamp:   header.Timestamp,
				Content: &nebulav1.TransactionReceipt_CancelR{
					CancelR: &nebulav1.CancelOrderReceipt{},
				},
			})
		case nebulav1.TxType_SignProxyTx:
			receipts.Receipts = append(receipts.Receipts, &nebulav1.TransactionReceipt{
				Hash:        wtx.Hash().Bytes(),
				TxType:      tx.TxType,
				Success:     true,
				BlockHeight: header.Height,
				Timestamp:   header.Timestamp,
				Content: &nebulav1.TransactionReceipt_SignProxyR{
					SignProxyR: &nebulav1.SignProxyReceipt{},
				},
			})
		case nebulav1.TxType_CreatePairTx:
			receipts.Receipts = append(receipts.Receipts, &nebulav1.TransactionReceipt{
				Hash:        wtx.Hash().Bytes(),
				TxType:      tx.TxType,
				Success:     true,
				BlockHeight: header.Height,
				Timestamp:   header.Timestamp,
				Content: &nebulav1.TransactionReceipt_CreatePairR{
					CreatePairR: &nebulav1.CreatePairReceipt{},
				},
			})
		case nebulav1.TxType_DisablePairTx:
			receipts.Receipts = append(receipts.Receipts, &nebulav1.TransactionReceipt{
				Hash:        wtx.Hash().Bytes(),
				TxType:      tx.TxType,
				Success:     true,
				BlockHeight: header.Height,
				Timestamp:   header.Timestamp,
				Content: &nebulav1.TransactionReceipt_DisablePairR{
					DisablePairR: &nebulav1.DisablePairReceipt{},
				},
			})
		default:
			panic("not implemented")
		}
	}

	return txs, receipts, nil
}

func (e *Engine) ProcessTx(header *nebulav1.BlockHeader, txs *nebulav1.TransactionList) (*nebulav1.TransactionReceiptList, error) {
	receipts := &nebulav1.TransactionReceiptList{
		Receipts: make([]*nebulav1.TransactionReceipt, 0),
	}
	for _, tx := range txs.Txs {
		wtx := wrapper.NewTxWrapper(tx)
		receipt := &nebulav1.TransactionReceipt{
			Hash:        wtx.Hash().Bytes(),
			TxType:      tx.TxType,
			Success:     true,
			BlockHeight: header.Height,
			Timestamp:   header.Timestamp,
		}
		switch tx.TxType {
		case nebulav1.TxType_DepositTx:
			address := common.Address(tx.Tx.(*nebulav1.Transaction_DepositTx).DepositTx.User)
			coin := types.Coin(tx.Tx.(*nebulav1.Transaction_DepositTx).DepositTx.Coin)
			amount := tx.Tx.(*nebulav1.Transaction_DepositTx).DepositTx.Amount
			amt := new(uint256.Int).SetBytes(amount)
			err := e.ProcessDepositBalance(address, coin, amt)
			if err != nil {
				receipt.Success = false
			}
			receipt.Content = &nebulav1.TransactionReceipt_DepositR{
				DepositR: &nebulav1.DepositReceipt{},
			}
		case nebulav1.TxType_CreatePairTx:
			err := e.ProcessCreatePair(tx.GetCreatePairTx())
			if err != nil {
				receipt.Success = false
			}
			receipt.Content = &nebulav1.TransactionReceipt_CreatePairR{
				CreatePairR: &nebulav1.CreatePairReceipt{},
			}
		case nebulav1.TxType_DisablePairTx:
			err := e.ProcessDisablePair(tx.GetDisablePairTx())
			if err != nil {
				receipt.Success = false
			}
			receipt.Content = &nebulav1.TransactionReceipt_DisablePairR{
				DisablePairR: &nebulav1.DisablePairReceipt{},
			}
		case nebulav1.TxType_MarketTx:
			coins := strings.Split(tx.GetMarketTx().Pair, "-")
			baseCoin := types.Coin(coins[0])
			quoteCoin := types.Coin(coins[1])
			user := common.HexToAddress(tx.User)
			side := orderbook.Buy
			if tx.GetMarketTx().Side == nebulav1.OrderSide_SELL {
				side = orderbook.Sell
			}
			quantity := new(uint256.Int).SetBytes(tx.GetMarketTx().Quantity)
			nonce := new(big.Int).SetBytes(tx.Nonce)
			_, _, err := e.ProcessMarketOrder(tx.GetMarketTx().OrderId,
				baseCoin, quoteCoin, user, side, quantity, nonce.Uint64(),
				tx.Proxy,
				utils.CombineRSV(tx.Signature.R, tx.Signature.S, uint8(tx.Signature.V)))
			if err != nil {
				receipt.Success = false
			}
			receipt.Content = &nebulav1.TransactionReceipt_MarketR{
				MarketR: &nebulav1.MarketOrderReceipt{},
			}
		case nebulav1.TxType_LimitTx:
			coins := strings.Split(tx.GetLimitTx().Pair, "-")
			baseCoin := types.Coin(coins[0])
			quoteCoin := types.Coin(coins[1])
			user := common.HexToAddress(tx.User)
			side := orderbook.Buy
			if tx.GetLimitTx().Side == nebulav1.OrderSide_SELL {
				side = orderbook.Sell
			}
			quantity := new(uint256.Int).SetBytes(tx.GetLimitTx().Quantity)
			price := new(uint256.Int).SetBytes(tx.GetLimitTx().Price)
			nonce := new(big.Int).SetBytes(tx.Nonce)
			_, _, err := e.ProcessLimitOrder(tx.GetLimitTx().OrderId,
				baseCoin, quoteCoin, user, side, quantity, price, nonce.Uint64(),
				tx.Proxy,
				utils.CombineRSV(tx.Signature.R, tx.Signature.S, uint8(tx.Signature.V)))
			if err != nil {
				receipt.Success = false
			}
			receipt.Content = &nebulav1.TransactionReceipt_LimitR{
				LimitR: &nebulav1.LimitOrderReceipt{},
			}
		case nebulav1.TxType_CancelTx:
			coins := strings.Split(tx.GetCancelTx().Pair, "-")
			baseCoin := types.Coin(coins[0])
			quoteCoin := types.Coin(coins[1])
			user := common.HexToAddress(tx.User)
			err := e.ProcessCancelOrder(tx.GetCancelTx().OrderId, baseCoin, quoteCoin, user, new(big.Int).SetBytes(tx.Nonce).Uint64(), tx.Proxy, utils.CombineRSV(tx.Signature.R, tx.Signature.S, uint8(tx.Signature.V)))
			if err != nil {
				receipt.Success = false
			}
			receipt.Content = &nebulav1.TransactionReceipt_CancelR{
				CancelR: &nebulav1.CancelOrderReceipt{},
			}
		default:
		}
		receipts.Receipts = append(receipts.Receipts, receipt)
	}
	return receipts, nil
}
